<?php

namespace App\Http\Controllers;

use App\Helpers\TimeLine;
use App\Models\Appointment;
use App\Models\BDTDevice;
use App\Models\CareMonth;
use App\Models\Client;
use App\Models\ClientInfoLine;
use App\Models\Facility;
use App\Models\NoteTemplate;
use App\Models\Pro;
use App\Models\ProGeneralAvailability;
use App\Models\ProSpecificAvailability;
use App\Models\ProSpecificUnavailability;
use App\Models\SectionTemplate;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\File;

class AppointmentController extends Controller
{
    public function events(Request $request)
    {
        $proIds = explode(',', $request->get('proIds'));
        $start = $request->get('start');
        $end = $request->get('end');
        $clientId = $request->get('clientId');
        $timeZone = $request->get('timeZone');

        // get appointments
        $appointments = Appointment
            ::where('status', '!=', 'COMPLETED')
            ->where('status', '!=', 'CANCELLED')
            ->where('start_time', '>=', $start)
            ->where('start_time', '<=', $end)
            ->where(function ($query) use ($proIds, $clientId) {
                $query
                    ->whereIn('pro_id', $proIds)
                    ->orWhere('client_id', '=', $clientId);
            })
            ->get();
        $events = [];
        foreach ($appointments as $appointment) {
            $events[] = [
                "type" => "appointment",
                "title" => ($appointment->client->id != $clientId ? '* ' : '') . $appointment->pro->displayName() . " (booked)",
                "_title" => $appointment->title,
                "description" => $appointment->description,
                "clientName" => $appointment->client->displayName(),
                "appointmentUid" => $appointment->uid,
                "clientUid" => $appointment->client->uid,
                "proId" => $appointment->pro->id,
                "proUid" => $appointment->pro->uid,
                "start" => $this->convertToTimezone($appointment->start_time, $timeZone),
                "end" => $this->convertToTimezone($appointment->end_time, $timeZone),
                "clientOnly" => !in_array($appointment->pro->id, $proIds),
                "otherClient" => ($appointment->client->id != $clientId),
                "editable" => true
            ];
        }

        // get availability
        $genAvail = ProGeneralAvailability
            ::where('is_cancelled', false)
            ->whereIn('pro_id', $proIds)
            ->get();
        $specAvail = ProSpecificAvailability
            ::where('is_cancelled', false)
            ->whereIn('pro_id', $proIds)
            ->where(function ($query) use ($start, $end) {
                $query
                    ->where(function ($query2) use ($start, $end) {
                        $query2
                            ->where('start_time', '>=', $start)
                            ->where('start_time', '<=', $end);
                    })
                    ->orWhere(function ($query2) use ($start, $end) {
                        $query2
                            ->where('end_time', '>=', $start)
                            ->where('end_time', '<=', $end);
                    });
            })
            ->get();
        $specUnavail = ProSpecificUnavailability
            ::where('is_cancelled', false)
            ->whereIn('pro_id', $proIds)
            ->where(function ($query) use ($start, $end) {
                $query
                    ->where(function ($query2) use ($start, $end) {
                        $query2
                            ->where('start_time', '>=', $start)
                            ->where('start_time', '<=', $end);
                    })
                    ->orWhere(function ($query2) use ($start, $end) {
                        $query2
                            ->where('end_time', '>=', $start)
                            ->where('end_time', '<=', $end);
                    });
            })
            ->get();

        // logic
        // 1. enumerate days between start and end (inclusive)
            // 2. for each pro
                // 3. calculate pairs of start-end of availability

        // 1. enumerate days between start and end (inclusive)
        $phpTZ = $this->appTZtoPHPTZ($timeZone);
        $startDate = new \DateTime($start, new \DateTimeZone($phpTZ));
        $endDate = new \DateTime($end, new \DateTimeZone($phpTZ));
        $endDate->setTime(23,59,59);
        $period = new \DatePeriod($startDate, \DateInterval::createFromDateString('1 day'), $endDate);
        $days = [];
        foreach ($period as $day) {
            $days[] = [
                "day" => strtoupper($day->format("l")), // SUNDAY, etc.
                "date" => $day->format("Y-m-d"), // 2020-10-04, etc.
            ];
        }

        foreach ($proIds as $proId) {

            $proTimeLine = new TimeLine($startDate, $endDate);

            $pro = Pro::where('id', $proId)->first();

            $proGenAvail = $genAvail->filter(function ($record) use ($proId) {
                return $record->pro_id == $proId;
            });

            // if no gen avail, assume mon to fri, 9 to 7
            if(!count($proGenAvail)) {
                $dayNames = ['MONDAY', 'TUESDAY', 'WEDNESDAY', 'THURSDAY', 'FRIDAY'];
                foreach ($dayNames as $dayName) {
                    $item = new \stdClass();
                    $item->day_of_week = $dayName;
                    $item->timezone = $timeZone;
                    $item->start_time = '09:00:00';
                    $item->end_time = '18:00:00';
                    $proGenAvail->push($item);
                }
            }

            $proSpecAvail = $specAvail->filter(function ($record) use ($proId) {
                return $record->pro_id == $proId;
            });

            $proSpecUnavail = $specUnavail->filter(function ($record) use ($proId) {
                return $record->pro_id == $proId;
            });

            // general availability
            foreach ($days as $day) {

                $proGenAvailForTheDay = $proGenAvail->filter(function ($record) use ($day) {
                    return $record->day_of_week === $day["day"];
                });
                foreach ($proGenAvailForTheDay as $ga) {

                    $gaStart = new \DateTime($day["date"], new \DateTimeZone($this->appTZtoPHPTZ($ga->timezone)));
                    $parts = explode(":", $ga->start_time);
                    $gaStart->setTime(intval($parts[0]), intval($parts[1]), intval($parts[2]));
                    $gaStart->setTimezone(new \DateTimeZone($phpTZ));

                    $gaEnd = new \DateTime($day["date"], new \DateTimeZone($this->appTZtoPHPTZ($ga->timezone)));
                    $parts = explode(":", $ga->end_time);
                    $gaEnd->setTime(intval($parts[0]), intval($parts[1]), intval($parts[2]));
                    $gaEnd->setTimezone(new \DateTimeZone($phpTZ));

                    $proTimeLine->addAvailability($gaStart, $gaEnd);
                }

            }

            // specific availability
            foreach ($proSpecAvail as $sa) {
                $saStart = new \DateTime($sa->start_time, new \DateTimeZone($this->appTZtoPHPTZ($sa->timezone)));
                $saStart->setTimezone(new \DateTimeZone($phpTZ));
                $saEnd = new \DateTime($sa->end_time, new \DateTimeZone($this->appTZtoPHPTZ($sa->timezone)));
                $saEnd->setTimezone(new \DateTimeZone($phpTZ));
                $proTimeLine->addAvailability($saStart, $saEnd);
            }

            // specific unavailability
            foreach ($proSpecUnavail as $sua) {
                $suaStart = new \DateTime($sua->start_time, new \DateTimeZone($this->appTZtoPHPTZ($sua->timezone)));
                $suaStart->setTimezone(new \DateTimeZone($phpTZ));
                $suaEnd = new \DateTime($sua->end_time, new \DateTimeZone($this->appTZtoPHPTZ($sua->timezone)));
                $suaEnd->setTimezone(new \DateTimeZone($phpTZ));
                $proTimeLine->removeAvailability($suaStart, $suaEnd);
            }

            // make already booked slots unavailable
            $proAppointments = $appointments->filter(function ($record) use ($proId) {
                return $record->pro_id == $proId;
            });
            foreach ($proAppointments as $appointment) {
                if($appointment->start_time && $appointment->end_time) {
                    $appStart = $this->convertToTimezone($appointment->start_time, $timeZone, 'UTC', true);
                    $appEnd = $this->convertToTimezone($appointment->end_time, $timeZone, 'UTC', true);
                    $proTimeLine->removeAvailability($appStart, $appEnd);
                }
            }

            foreach ($proTimeLine->available as $item) {

                $eStart = new \DateTime('@' . $item->start);
                $eStart->setTimezone(new \DateTimeZone($phpTZ));

                $eEnd = new \DateTime('@' . $item->end);
                $eEnd->setTimezone(new \DateTimeZone($phpTZ));

                $events[] = [
                    "type" => "availability",
                    "title" => $pro->displayName() . " (available)",
                    "proId" => $pro->id,
                    "proUid" => $pro->uid,
                    "start" => $eStart->format('Y-m-d H:i:s'),
                    "end" => $eEnd->format('Y-m-d H:i:s'),
                    "editable" => false
                ];

            }

        }

        return json_encode($events);
    }

    private function convertToTimezone($_dateTime, $_targetTimezone, $_sourceTimezone = 'UTC', $_returnRaw = false) {
        if(!$_dateTime) return $_dateTime;
        $date = new \DateTime($_dateTime, new \DateTimeZone($_sourceTimezone));
        $date->setTimezone(new \DateTimeZone($this->appTZtoPHPTZ($_targetTimezone)));
        return $_returnRaw ? $date : $date->format('Y-m-d H:i:s');
    }

    private function appTZtoPHPTZ($_timezone) {
        switch($_timezone) {
            case 'ALASKA':
                $timezone = "US/Alaska";
                break;
            case 'CENTRAL':
                $timezone = "US/Central";
                break;
            case 'HAWAII':
                $timezone = "US/Hawaii";
                break;
            case 'MOUNTAIN':
                $timezone = "US/Mountain";
                break;
            case 'PACIFIC':
                $timezone = "US/Pacific";
                break;
            case 'PUERTO_RICO':
                $timezone = "America/Puerto_Rico";
                break;
            default:
                $timezone = "US/Eastern";
                break;
        }
        return $timezone;
    }
}
